#import "KochavaTrackerPlugin.h"
#import "KochavaTracker.h"

// Declare delegate for Kochava Attribution callback with the plugin.
@interface KochavaTrackerPlugin () <KochavaTrackerDelegate>
@end

@implementation KochavaTrackerPlugin

// Single dispatch for wrapper product registration.
static dispatch_once_t wrapper_product_dispatch_once_t = 0;

// Register plugin with Flutter.
+ (void)registerWithRegistrar:(NSObject <FlutterPluginRegistrar> *)registrar {
    FlutterMethodChannel *channel = [FlutterMethodChannel methodChannelWithName:@"kochava_tracker" binaryMessenger:[registrar messenger]];
    KochavaTrackerPlugin *instance = [[KochavaTrackerPlugin alloc] initWithChannel:channel];
    [registrar addMethodCallDelegate:instance channel:channel];
}

// Attempts to read an NSDictionary and returns nil if not one.
+ (NSDictionary *)readNSDictionary:(id)valueId {
    return [valueId isKindOfClass:NSDictionary.class] ? (NSDictionary *) valueId : nil;
}

// Attempts to read an NSString and returns nil if not one.
+ (NSString *)readNSString:(id)valueId {
    return [valueId isKindOfClass:NSString.class] ? (NSString *) valueId : nil;
}

// Attempts to read an NSNumber and returns nil if not one.
+ (NSNumber *)readNSNumber:(id)valueId {
    return [valueId isKindOfClass:NSNumber.class] ? (NSNumber *) valueId : nil;
}

// Returns the hex formatted NSString as NSData. Returns nil if unable.
// Discussion:  This is being employed to take the output of an NSData description (which is a hex string, such as is the case with a push notification token) and turn it back into an NSData.  This was sourced from the web and then optimized.
+ (id)dataWithHexString:(NSString *)hexString {
    // VALIDATION (RETURN)
    if (hexString == nil) {
        return nil;
    }

    // hexStringLength must be an even number of digits
    NSUInteger hexStringLength = hexString.length;
    if (hexStringLength % 2 > 0) {
        NSLog(@"Warning:  func dataWithHexString(_:) - parameter hexString was passed a value which does not have an even number of digits.  hexString = %@", hexString);
        return nil;
    }

    // Must contain only valid HEX characters.
    NSCharacterSet *chars = [[NSCharacterSet characterSetWithCharactersInString:@"0123456789ABCDEFabcdef"] invertedSet];
    BOOL isValid = (NSNotFound == [hexString rangeOfCharacterFromSet:chars].location);
    if (!isValid) {
        NSLog(@"Warning:  func dataWithHexString(_:) - parameter hexString was passed a value which was not valid HEX. hexString = %@", hexString);
        return nil;
    }

    // bytes and bytesPointer default to point to some newly allocated memory
    unsigned char *bytes = (unsigned char *) malloc(hexStringLength / 2);
    unsigned char *bytesPointer = bytes;

    // fill with long(s) converted from two-digit strings containing base-16 representations of numbers
    for (NSUInteger index = 0; index < hexStringLength; index += 2) {
        // buffer set to the two-digit base 16 number located at index
        char buffer[3];
        buffer[0] = (char) [hexString characterAtIndex:index];
        buffer[1] = (char) [hexString characterAtIndex:index + 1];
        buffer[2] = '\0';

        // longInt and endPointer set longInt to buffer converted to a long, and set endPointer to the next character in buffer after the numerical value.
        char *endPointer = NULL;
        long int longInt = strtol(buffer, &endPointer, 16);

        // bytesPointer update with longInt
        *bytesPointer = (unsigned char) longInt;

        // advance to next position
        bytesPointer++;
    }

    return [NSData dataWithBytesNoCopy:bytes length:(hexStringLength / 2) freeWhenDone:YES];
}

// Decodes the attribution dictionary from the callback delegate or getter and returns a string.
+ (NSString *)decodeAttributionDictionary:(NSDictionary *)attributionDictionary {
    // Ensure it is json.
    if (attributionDictionary == nil || ![NSJSONSerialization isValidJSONObject:attributionDictionary]) {
        return @"";
    }

    return [KochavaTrackerPlugin serializeNSDictionary:attributionDictionary];
}

// Serialize an NSDictionary into an NSString
+ (NSString *)serializeNSDictionary:(NSDictionary *)dictionary {
    NSError *error = nil;
    NSData *dictionaryJSONData = [NSJSONSerialization dataWithJSONObject:dictionary options:0 error:&error];
    if (dictionaryJSONData == nil) {
        return @"";
    }

    NSString *result = [[NSString alloc] initWithData:dictionaryJSONData encoding:NSUTF8StringEncoding];
    if (result == nil) {
        return @"";
    }

    return result;
}

// Clears the running instance of the KochavaTracker so it is no longer running.
+ (void)invalidateKochava {
    [KochavaTracker.shared performSelector:@selector(invalidate)];
}

// Removes all saved state for the KochavaTracer. Should call invalidate first.
+ (void)removeKochavaUserDefaults {
    // Remove the NSUserDefaults keys.
    NSArray *keyArray = NSUserDefaults.standardUserDefaults.dictionaryRepresentation.allKeys;
    for (id key in keyArray) {
        NSRange kochavaPrefixRange = [key rangeOfString:@"com.kochava"];
        if (kochavaPrefixRange.location != NSNotFound) {
            [NSUserDefaults.standardUserDefaults removeObjectForKey:key];
        }
    }

    // Remove the deviceId backup file.
    NSURL *documentDirectoryURL = [NSFileManager.defaultManager URLsForDirectory:NSApplicationSupportDirectory inDomains:NSUserDomainMask][0];
    NSURL *kochavaDirectoryURL = [documentDirectoryURL URLByAppendingPathComponent:@"com.kochava.KochavaTracker" isDirectory:true];
    NSURL *backupURL = [kochavaDirectoryURL URLByAppendingPathComponent:@"deviceIdString"];

    NSError *error = nil;
    [NSFileManager.defaultManager removeItemAtURL:backupURL error:&error];
}

// Initialize the plugin with the method channel to communicate with Dart.
- (instancetype)initWithChannel:(FlutterMethodChannel *)channel {
    self = [super init];
    if (self) {
        _channel = channel;
    }
    return self;
}

// Handle a method call from the Dart layer.
- (void)handleMethodCall:(FlutterMethodCall *)call result:(FlutterResult)result {
    if ([@"configure" isEqualToString:call.method]) {
        NSDictionary *valueDictionary = [KochavaTrackerPlugin readNSDictionary:call.arguments];
        if (valueDictionary == nil) {
            result(@"Failed");
            return;
        }
        [self configure:valueDictionary.mutableCopy];
        result(@"success");

    } else if ([@"getAttribution" isEqualToString:call.method]) {
        NSDictionary *attributionDictionary = KochavaTracker.shared.attributionDictionary;
        NSString *attributionString = [KochavaTrackerPlugin decodeAttributionDictionary:attributionDictionary];
        result(attributionString);

    } else if ([@"getDeviceId" isEqualToString:call.method]) {
        result(KochavaTracker.shared.deviceIdString ?: @"");

    } else if ([@"getVersion" isEqualToString:call.method]) {
        result(KochavaTracker.shared.sdkVersionString ?: @"");

    } else if ([@"setAppLimitAdTracking" isEqualToString:call.method]) {
        NSNumber *valueNumber = [KochavaTrackerPlugin readNSNumber:call.arguments];
        if (valueNumber == nil) {
            result(@"Failed");
            return;
        }
        BOOL value = [valueNumber boolValue];
        [KochavaTracker.shared setAppLimitAdTrackingBool:value];
        result(@"success");

    } else if ([@"setIdentityLink" isEqualToString:call.method]) {
        NSDictionary *valueDictionary = [KochavaTrackerPlugin readNSDictionary:call.arguments];
        if (valueDictionary == nil) {
            result(@"Failed");
            return;
        }
        [KochavaTracker.shared sendIdentityLinkWithDictionary:valueDictionary];
        result(@"success");

    } else if ([@"setSleep" isEqualToString:call.method]) {
        NSNumber *valueNumber = [KochavaTrackerPlugin readNSNumber:call.arguments];
        if (valueNumber == nil) {
            result(@"Failed");
            return;
        }
        BOOL value = [valueNumber boolValue];
        [KochavaTracker.shared setSleepBool:value];
        result(@"success");

    } else if ([@"getSleep" isEqualToString:call.method]) {
        result(@(KochavaTracker.shared.sleepBool));

    } else if ([@"addPushToken" isEqualToString:call.method]) {
        NSString *valueString = [KochavaTrackerPlugin readNSString:call.arguments];
        if(valueString.length == 0) {
            result(@"Failed");
            return;
        }
        NSData *tokenData = [self.class dataWithHexString:valueString];
        if (tokenData == nil) {
            result(@"Failed");
            return;
        }
        [KochavaTracker.shared addRemoteNotificationsDeviceToken:tokenData];
        result(@"success");

    } else if ([@"removePushToken" isEqualToString:call.method]) {
        NSString *valueString = [KochavaTrackerPlugin readNSString:call.arguments];
        if(valueString.length == 0) {
            result(@"Failed");
            return;
        }
        NSData *tokenData = [self.class dataWithHexString:valueString];
        [KochavaTracker.shared removeRemoteNotificationsDeviceToken:tokenData];
        result(@"success");

    } else if ([@"sendEventString" isEqualToString:call.method]) {
        NSDictionary *valueDictionary = [KochavaTrackerPlugin readNSDictionary:call.arguments];
        if (valueDictionary == nil) {
            result(@"Failed");
            return;
        }
        NSString *nameString = [KochavaTrackerPlugin readNSString:valueDictionary[@"name"]];
        NSString *infoString = [KochavaTrackerPlugin readNSString:valueDictionary[@"info"]];
        if(nameString.length == 0) {
            result(@"Failed");
            return;
        }
        [KochavaTracker.shared sendEventWithNameString:nameString infoString:infoString];
        result(@"success");

    } else if ([@"sendEventMapObject" isEqualToString:call.method]) {
        NSDictionary *valueDictionary = [KochavaTrackerPlugin readNSDictionary:call.arguments];
        if (valueDictionary == nil) {
            result(@"Failed");
            return;
        }
        NSString *nameString = [KochavaTrackerPlugin readNSString:valueDictionary[@"name"]];
        NSDictionary *infoDictionary = [KochavaTrackerPlugin readNSDictionary:valueDictionary[@"info"]];
        if(nameString.length == 0) {
            result(@"Failed");
            return;
        }
        [KochavaTracker.shared sendEventWithNameString:nameString infoDictionary:infoDictionary];
        result(@"success");

    } else if ([@"sendEventAppleAppStoreReceipt" isEqualToString:call.method]) {
        NSDictionary *valueDictionary = [KochavaTrackerPlugin readNSDictionary:call.arguments];
        if (valueDictionary == nil) {
            result(@"Failed");
            return;
        }
        NSString *nameString = [KochavaTrackerPlugin readNSString:valueDictionary[@"name"]];
        NSDictionary *infoDictionary = [KochavaTrackerPlugin readNSDictionary:valueDictionary[@"info"]];
        NSString *appStoreReceiptBase64EncodedString = [KochavaTrackerPlugin readNSString:valueDictionary[@"appStoreReceiptBase64EncodedString"]];
        if(nameString.length == 0) {
            result(@"Failed");
            return;
        }

        KochavaEvent *event = [KochavaEvent eventWithEventTypeEnum:KochavaEventTypeEnumCustom];
        event.customEventNameString = nameString;
        if (infoDictionary != nil) {
            event.infoDictionary = infoDictionary;
        }
        if (appStoreReceiptBase64EncodedString.length > 0) {
            event.appStoreReceiptBase64EncodedString = appStoreReceiptBase64EncodedString;
        }
        [KochavaTracker.shared sendEvent:event];
        result(@"success");

    } else if ([@"sendEventGooglePlayReceipt" isEqualToString:call.method]) {
        // Not supported on this platform.
        result(@"success");

    } else if ([@"sendDeepLink" isEqualToString:call.method]) {
        NSDictionary *valueDictionary = [KochavaTrackerPlugin readNSDictionary:call.arguments];
        if (valueDictionary == nil) {
            result(@"Failed");
            return;
        }
        NSString *openURLString = [KochavaTrackerPlugin readNSString:valueDictionary[@"openURLString"]];
        NSString *sourceApplicationString = [KochavaTrackerPlugin readNSString:valueDictionary[@"sourceApplicationString"]];

        // Create and send the deeplink event.
        KochavaEvent *event = [KochavaEvent eventWithEventTypeEnum:KochavaEventTypeEnumDeepLink];
        if (openURLString.length > 0) {
            event.uriString = openURLString;
        }
        if (sourceApplicationString.length > 0) {
            event.sourceString = sourceApplicationString;
        }
        [KochavaTracker.shared sendEvent:event];
        result(@"success");

    } else if ([@"processDeeplink" isEqualToString:call.method]) {
        NSDictionary *valueDictionary = [KochavaTrackerPlugin readNSDictionary:call.arguments];
        if (valueDictionary == nil) {
            result(@"Failed");
            return;
        }
        NSString *id = [KochavaTrackerPlugin readNSString:valueDictionary[@"id"]];
        NSString *path = [KochavaTrackerPlugin readNSString:valueDictionary[@"path"]];
        NSURL *pathUrl = nil;
        if (path.length > 0) {
            pathUrl = [NSURL URLWithString:path];
        }
        NSNumber *timeout = [KochavaTrackerPlugin readNSNumber:valueDictionary[@"timeout"]];
        NSTimeInterval timeoutTimeInterval = 10.0;
        if (timeout != nil) {
            timeoutTimeInterval = timeout.doubleValue;
        }

        // Process the deeplink.
        [KVADeeplink processWithURL:pathUrl timeoutTimeInterval:timeoutTimeInterval completionHandler:^(KVADeeplink *_Nonnull deeplink) {
            // Serialize the response deeplink inside a dictionary with the request id.
            NSObject *deeplinkAsForContextObject = [deeplink kva_asForContextObjectWithContext:KVAContext.sdkWrapper];
            NSDictionary *deeplinkDictionary = [deeplinkAsForContextObject isKindOfClass:NSDictionary.class] ? (NSDictionary *) deeplinkAsForContextObject : nil;

            NSMutableDictionary *responseDictionary = [[NSDictionary alloc] init].mutableCopy;
            responseDictionary[@"id"] = id;
            responseDictionary[@"deeplink"] = deeplinkDictionary;

            NSString *response = [KochavaTrackerPlugin serializeNSDictionary:responseDictionary];
            if (response.length == 0) {
                response = @"{}";
            }

            // Emit the deeplink event notification.
            [self.channel invokeMethod:@"deeplinkCallback" arguments:response];
        }];
        result(@"success");

    } else {
        result(FlutterMethodNotImplemented);
    }
}

// Handle retrieving the attribution callback and sending it to the Dart layer.
- (void)tracker:(nonnull KochavaTracker *)tracker didRetrieveAttributionDictionary:(nonnull NSDictionary *)attributionDictionary {
    NSString *attributionString = [KochavaTrackerPlugin decodeAttributionDictionary:attributionDictionary];
    [self.channel invokeMethod:@"attributionCallback" arguments:attributionString];
}

// Configure and start the SDK.
- (void)configure:(NSMutableDictionary *)trackerParametersDictionary {
    // Check for the existence of the hidden unconfigure key.
    if (trackerParametersDictionary[@"INTERNAL_UNCONFIGURE"]) {
        NSLog(@"KochavaTrackerPlugin.configure UnConfigure.");
        [KochavaTrackerPlugin invalidateKochava];
        return;
    }

    // Check for the existence of the hidden reset key.
    if (trackerParametersDictionary[@"INTERNAL_RESET"]) {
        NSLog(@"KochavaTrackerPlugin.configure Reset.");
        [KochavaTrackerPlugin removeKochavaUserDefaults];
        return;
    }

    // Retrieve and set the App Guid
    id appGUIDStringObject = trackerParametersDictionary[@"iOSAppGUIDString"];
    if (appGUIDStringObject != nil) {
        trackerParametersDictionary[kKVAParamAppGUIDStringKey] = appGUIDStringObject;
        trackerParametersDictionary[@"iOSAppGUIDString"] = nil;
    }

    // Wrapper Extension info
    dispatch_once(&wrapper_product_dispatch_once_t, ^{
        NSString *wrapperNameString = trackerParametersDictionary[@"wrapperNameString"];
        NSString *wrapperVersionString = trackerParametersDictionary[@"wrapperVersionString"];
        NSString *wrapperBuildDateString = trackerParametersDictionary[@"wrapperBuildDateString"];

        // wrapperProduct Configure and register.
        KVAProduct *wrapperProduct = [KVAProduct
                productWithAPIVersionString:@"3"
                buildDateString:wrapperBuildDateString
                bundleIdentifierString:nil
                bundleTypeString:@"Wrapper"
                compilerFlagNameStringArray:nil
                compilerFlagPredicateSubstitutionVariablesDictionary:nil
                moduleNameString:@"KochavaFlutter"
                nameString:wrapperNameString
                organizationNameString:@"Kochava"
                reverseDomainNameString:@"com.kochava.flutter"
                versionString:wrapperVersionString
        ];

        [wrapperProduct register];

        // KVAProduct(s) Configure.  Also register any which are optional.
        KVACoreProduct.shared.wrapperProduct = wrapperProduct;
        KVATrackerProduct.shared.wrapperProduct = wrapperProduct;
    });
    trackerParametersDictionary[@"wrapperNameString"] = nil;
    trackerParametersDictionary[@"wrapperVersionString"] = nil;
    trackerParametersDictionary[@"wrapperBuildDateString"] = nil;

    // Retrieve sleep info to set after starting the SDK.
    BOOL sleepBool = [trackerParametersDictionary[@"sleepBool"] boolValue];
    trackerParametersDictionary[@"sleepBool"] = nil;

    // Start the SDK.
    [KochavaTracker.shared configureWithParametersDictionary:trackerParametersDictionary delegate:self];

    //Check for sleep.
    if (sleepBool) {
        [KochavaTracker.shared setSleepBool:sleepBool];
    }
}

@end
